"""
Keypair management
"""
import base64
import hashlib

from novaclient.client import Client as NovaClient
from novaclient.exceptions import NotFound

from . import context


def ssh_fingerprint(key):
    key = base64.b64decode(key.split()[1].encode("ascii"))
    fingerprint = hashlib.md5(key).hexdigest()
    return fingerprint


def key_pair_name(fingerprint):
    return "keypair-{}".format(fingerprint)


class Keypair(object):
    def __init__(self, **kwargs):
        kwargs.setdefault("session", context.session())

        key_filename = kwargs.get(
            "keypair_public_key", context.get("keypair_public_key")
        )
        session = kwargs.get("session")

        self.nova = NovaClient("2", session=session)

        with open(key_filename) as f:
            self.key = [line.strip() for line in f.readlines() if line.strip()][0]
            fingerprint = ssh_fingerprint(self.key)
            self.key_name = key_pair_name(fingerprint)

        try:
            self.key_pair = self.nova.keypairs.get(self.key_name)
        except NotFound:
            self.key_pair = self.nova.keypairs.create(
                self.key_name, public_key=self.key
            )



from datetime import datetime
import pytest

from chi.server import BAREMETAL_FLAVOR, DEFAULT_IMAGE, DEFAULT_NETWORK
from collections import namedtuple
from contextlib import nullcontext


@pytest.fixture()
def now():
    return datetime(2021, 1, 1, 0, 0, 0, 0)


def example_create_server():
    """Launch a bare metal instance.

    <div class="alert alert-info">

    **Functions used in this example:**

    * [create_server](../modules/server.html#chi.server.create_server)
    * [get_node_reservation](../modules/lease.html#chi.lease.get_node_reservation)

    </div>

    """
    from chi.lease import get_node_reservation
    from chi.server import create_server

    # We assume a lease has already been created, for example with
    # ``chi.lease.create_lease```
    lease_name = "my_lease"
    server_name = "my_server"
    reservation_id = get_node_reservation(lease_name)
    server = create_server(server_name, reservation_id=reservation_id)


def test_example_create_server(mocker):
    nova = mocker.patch("chi.server.nova")()

    def get_network_id(network_name):
        assert network_name == DEFAULT_NETWORK
        return "network-id"

    def get_image_id(image_name):
        assert image_name == DEFAULT_IMAGE
        return "image-id"

    def list_flavors():
        return [namedtuple("Flavor", ["id"])("flavor-id")]

    def update_keypair(key_name=None, public_key=None):
        return namedtuple("Keypair", ["id"])("fake-key")

    mocker.patch("chi.server.get_network_id", side_effect=get_network_id)
    mocker.patch("chi.server.get_image_id", side_effect=get_image_id)
    mocker.patch("chi.server.list_flavors", side_effect=list_flavors)
    mocker.patch("chi.lease.get_node_reservation",
                 return_value="reservation-id")
    mocker.patch("chi.server.update_keypair", side_effect=update_keypair)

    example_create_server()

    nova.servers.create.assert_called_once_with(
        name="my_server",
        flavor="flavor-id",
        image="image-id",
        key_name="fake-key",
        nics=[{"net-id": "network-id", "v4-fixed-ip": ""}],
        scheduler_hints={"reservation": "reservation-id"},
        max_count=1,
        min_count=1,
        hypervisor_hostname=None
    )


def example_wait_for_connectivity():
    """Wait for a server's port to come up before proceeding.

    Sometimes you want to interact with the server over a remote interface
    and need to wait until it's up and accepting connections. The
    :func:`~chi.server.wait_for_tcp` function allows you to do just that.
    This example also illustrates how you can bind a Floating IP (public IP)
    to the server so it can be reached over the internet.

    <div class="alert alert-info">

    **Functions used in this example:**

    * [associate_floating_ip](../modules/server.html#chi.server.associate_floating_ip)
    * [wait_for_tcp](../modules/server.html#chi.server.wait_for_tcp)

    </div>
    """
    from chi.server import associate_floating_ip, wait_for_tcp

    # Note: this is a placeholder server ID. Yours will be different!
    # server_id can be obtained like `server.id` if you created the server
    # with `create_server`. It can also be obtained via `get_server_id(name)`
    server_id = "6b2bae1e-0311-493f-836c-a9da0cb9e0c0"
    ip = associate_floating_ip(server_id)

    # Wait for SSH connectivity over port 22
    wait_for_tcp(ip, port=22)


def test_example_wait_for_connectivity(mocker):
    connection = mocker.patch("chi.server.connection")()

    def get_free_floating_ip():
        return {"floating_ip_address": "fake-floating-ip"}

    mocker.patch("chi.server.get_free_floating_ip",
                 side_effect=get_free_floating_ip)
    socket_create = mocker.patch("chi.server.socket.create_connection")
    socket_create.return_value = nullcontext()

    example_wait_for_connectivity()

    connection.compute.add_floating_ip_to_server.assert_called_once_with(
        "6b2bae1e-0311-493f-836c-a9da0cb9e0c0", "fake-floating-ip")
    socket_create.assert_called_once_with(
        ("fake-floating-ip", 22), timeout=(60 * 20))



from fabric import Connection
from paramiko.client import WarningPolicy

from . import context


class Remote(Connection):
    def __init__(self, ip=None, server=None, user="cc"):
        if ip is None:
            if server is None:
                raise ValueError("ip or server must be provided.")
            ip = server.ip

        key_filename = context.get("keypair_private_key")
        connect_kwargs = {"key_filename": key_filename}
        super(Remote, self).__init__(ip, user=user, connect_kwargs=connect_kwargs)
        # Default policy is to reject unknown hosts - for our use-case,
        # printing a warning is probably enough, given the host is almost
        # always guaranteed to be unknown.
        self.client.set_missing_host_key_policy(WarningPolicy)



from .clients import (
    connection,
    blazar,
    glance,
    gnocchi,
    ironic,
    keystone,
    neutron,
    nova,
    zun,
)
from .context import reset, set, get, params, session, use_site


__all__ = [
    "get",
    "params",
    "reset",
    "session",
    "set",
    "use_site",
    "connection",
    "blazar",
    "glance",
    "gnocchi",
    "ironic",
    "keystone",
    "neutron",
    "nova",
    "zun",
]



# DEPRECATED
# This module is an alias to the server module and should not be used.
# It is temporarily left in place so that existing references do not break.
from .server import *

print("""
The 'server_api_examples' module is deprecated! Use the 'server' module
directly going forward.
""")



# Copyright 2021 University of Chicago
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import namedtuple
from datetime import datetime
import pytest

from chi.container import DEFAULT_IMAGE_DRIVER, DEFAULT_NETWORK


@pytest.fixture()
def now():
    return datetime(2021, 1, 1, 0, 0, 0, 0)


def example_create_container():
    """Launch a container.

    <div class="alert alert-info">

    **Functions used in this example:**

    * [create_container](../modules/container.html#chi.container.create_container)
    * [get_device_reservation](../modules/lease.html#chi.lease.get_device_reservation)

    </div>

    """
    from chi.lease import get_device_reservation
    from chi.container import create_container

    # We assume a lease has already been created, for example with
    # ``chi.lease.create_lease```
    lease_name = "my_lease"
    container_name = "my_container"
    reservation_id = get_device_reservation(lease_name)
    container = create_container(
        container_name,
        image="centos:8",
        reservation_id=reservation_id,
    )


def test_example_create_container(mocker):
    zun = mocker.patch("chi.container.zun")()

    mocker.patch("chi.lease.get_device_reservation", return_value="reservation-id")
    Container = namedtuple("Container", ["uuid", "name", "status"])
    mocker.patch(
        # Fake that the container is already created
        "chi.container.get_container",
        return_value=Container("fake-uuid", "my-container", "Running"),
    )

    example_create_container()

    zun.containers.create.assert_called_once_with(
        name="my_container",
        image="centos:8",
        hints={"reservation": "reservation-id", "platform_version": 2},
    )



# Copyright 2021 University of Chicago
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import logging
import tarfile
import time
import typing

from .clients import zun
from .network import bind_floating_ip, get_free_floating_ip, get_network_id

if typing.TYPE_CHECKING:
    from zunclient.v1.containers import Container

DEFAULT_IMAGE_DRIVER = "docker"
DEFAULT_NETWORK = "containernet1"
LOG = logging.getLogger(__name__)

__all__ = [
    "list_containers",
    "get_container",
    "create_container",
    "snapshot_container",
    "destroy_container",
    "get_logs",
    "execute",
    "upload",
    "wait_for_active",
]


def create_container(
    name: "str",
    image: "str" = None,
    exposed_ports: "list[str]" = None,
    reservation_id: "str" = None,
    start: "bool" = True,
    start_timeout: "int" = None,
    platform_version: "int" = 2,
    **kwargs,
) -> "Container":
    """Create a container instance.

    Args:
        name (str): The name to give the container.
        image (str): The Docker image, with or without tag information. If no
            tag is provided, "latest" is assumed.
        device_profiles (list[str]): An optional list of device profiles to
            request be configured on the container when it is created. Edge
            devices may have differing sets of supported device profiles, so
            it is important to understand which profiles are supported by the
            target device for your container.
        environment (dict): A set of environment variables to pass to the
            container.
        exposed_ports (list[str]): A list of ports to expose on the container.
            TCP or UDP can be provided with a slash prefix, e.g., "80/tcp" vs.
            "53/udp". If no protocol is provided, TCP is assumed.
        host (str): The Zun host to launch a container on. If not specified,
            the host is chosen by Zun.
        runtime (str): The container runtime to use. This should only be
            overridden when explicitly launching containers onto a host/platform
            requiring a separate runtime to, e.g., pass-through GPU devices,
            such as the "nvidia" runtime provided by NVIDIA Jetson Nano/TX2.
        start (bool): Whether to automatically start the container after it
            is created. Default True.
        **kwargs: Additional keyword arguments to send to the Zun client's
            container create call.
    """
    hints = kwargs.setdefault("hints", {})
    if reservation_id:
        hints["reservation"] = reservation_id
    if platform_version:
        hints["platform_version"] = platform_version

    # Support simpler syntax for exposed_ports
    if exposed_ports and isinstance(exposed_ports, list):
        exposed_ports = {port_def: {} for port_def in exposed_ports}
    # Only set exposed_ports on the parent invocation if it is non-empty. Otherwise,
    # end-users cannot specify security groups; the client will send an explicit 'null'
    # value for this key, which will fail validation in the API layer, which expects the
    # key to be missing if security groups are specified.
    if exposed_ports:
        kwargs["exposed_ports"] = exposed_ports

    # Note: most documented args are not on the function signature because there is some special
    # handling of it in the Zun client; it is not sent if it is not on kwargs.
    # If it is on kwargs it is expected to be non-None.
    container = zun().containers.create(
        name=name,
        image=image,
        **kwargs,
    )

    # Wait for a while, the image may need to download. 30 minutes is
    # _quite_ a long time, but the user can interrupt or choose a smaller
    # timeout.
    timeout = start_timeout or (60 * 30)
    LOG.info(f"Waiting up to {timeout}s for container creation ...")

    if platform_version == 2:
        container = _wait_for_status(container.uuid, "Running", timeout=timeout)
    else:
        container = _wait_for_status(container.uuid, "Created", timeout=timeout)
        if start:
            LOG.info("Starting container ...")
            zun().containers.start(container.uuid)

    return container


def list_containers() -> "list[Container]":
    """List all containers owned by this project.

    Returns:
        A list of containers.
    """
    return zun().containers.list()


def get_container(container_ref: "str") -> "Container":
    """Get a container's information.

    Args:
        container_ref (str): The name or ID of the container.
        tag (str): An optional version to tag the container image with. If not
            defined, defaults to "latest".

    Returns:
        The container, if found.
    """
    return zun().containers.get(container_ref)


def snapshot_container(
    container_ref: "str", repository: "str", tag: "str" = "latest"
) -> "str":
    """Create a snapshot of a running container.

    This will store the container's file system in Glance as a new Image.
    You can then specify the Image ID in container create requests.

    Args:
        container_ref (str): The name or ID of the container.
        repository (str): The name to give the snapshot.
        tag (str): An optional version tag to give the snapshot. Defaults to
            "latest".
    """
    return zun().containers.commit(container_ref, repository, tag=tag)["uuid"]


def destroy_container(container_ref: "str"):
    """Delete the container.

    This will automatically stop the container if it is currently running.

    Args:
        container_ref (str): The name or ID of the container.
    """
    return zun().containers.delete(container_ref, stop=True)


def get_logs(container_ref: "str", stdout=True, stderr=True):
    """Print all logs outputted by the container.

    Args:
        container_ref (str): The name or ID of the container.
        stdout (bool): Whether to include stdout logs. Default True.
        stderr (bool): Whether to include stderr logs. Default True.

    Returns:
        A string containing all log output. Log lines will be delimited by
            newline characters.
    """
    return zun().containers.logs(container_ref, stdout=stdout, stderr=stderr)


def execute(container_ref: "str", command: "str") -> "dict":
    """Execute a one-off process inside a running container.

    Args:
        container_ref (str): The name or ID of the container.
        command (str): The command to run.

    Returns:
        A summary of the output of the command, with "output" and "exit_code".
    """
    return zun().containers.execute(container_ref, command=command, run=True)


def upload(container_ref: "str", source: "str", dest: "str") -> "dict":
    """Upload a file or directory to a running container.

    Args:
        container_ref (str): The name or ID of the container.
        source (str): The (local) path to the file or directory to upload.
        dest (str): The (container) path to upload the file or directory to.
    """
    fd = io.BytesIO()
    with tarfile.open(fileobj=fd, mode="w") as tar:
        tar.add(source, arcname=".")
    fd.seek(0)
    data = fd.read()
    fd.close()
    return zun().containers.put_archive(container_ref, dest, data)


def download(container_ref: "str", source: "str", dest: "str"):
    """Download a file or directory from a running container.

    Args:
        container_ref (str): The name or ID of the container.
        source (str): The (container) path of the file or directory.
        dest (str): The (local) path to download to.
    """
    res = zun().containers.get_archive(container_ref, source)
    fd = io.BytesIO(res["data"])
    with tarfile.open(fileobj=fd, mode="r") as tar:
        tar.extractall(dest)


def wait_for_active(container_ref: "str", timeout: int = (60 * 2)) -> "Container":
    """Wait for a container to transition to the running state.

    Args:
        container_ref (str): The name or ID of the container.
        timeout (int): How long to wait before issuing a TimeoutError.

    Raises:
        TimeoutError: if the timeout was reached before the container started.

    Returns:
        The container representation.
    """
    return _wait_for_status(container_ref, "Running", timeout=timeout)


def _wait_for_status(
    container_ref: "str", status: "str", timeout: int = (60 * 2)
) -> "Container":
    start_time = time.perf_counter()

    while True:
        container = get_container(container_ref)
        if container.status == "Error":
            raise RuntimeError("Container went in to error state")
        elif container.status == status:
            return container
        time.sleep(5)
        if time.perf_counter() - start_time >= timeout:
            raise TimeoutError(
                (
                    f"Waited too long for the container {container_ref} to be "
                    f"{status.lower()}."
                )
            )


def associate_floating_ip(container_ref: "str", floating_ip_address=None) -> "str":
    """Assign a Floating IP address to a container.

    The container's first address will be used for the assignment.

    Args:
        container_ref (str): The name or ID of the container.
        floating_ip_address (str): The Floating IP address, which must already
            be owned by the requesting project. If not defined, a Floating IP
            will be allocated, if there are any available.

    Returns:
        The Floating IP address, if it was bound successfully, else None.
    """
    if not floating_ip_address:
        floating_ip_address = get_free_floating_ip()["floating_ip_address"]

    container = zun().containers.get(container_ref)
    for net_id, addrs in container.addresses.items():
        port = next(iter([a["port"] for a in addrs if a["port"]]), None)
        if port:
            bind_floating_ip(floating_ip_address, port_id=port)
            return floating_ip_address

    return None



from datetime import timedelta
import json
import numbers
import re
import sys
import time
from typing import TYPE_CHECKING

from blazarclient.exception import BlazarClientException

from .clients import blazar, neutron
from .context import get as get_from_context, session
from .network import get_network_id, PUBLIC_NETWORK, list_floating_ips
from .server import Server, ServerError
from .util import random_base32, utcnow

import logging

if TYPE_CHECKING:
    from typing import Pattern

LOG = logging.getLogger(__name__)


class ErrorParsers:
    NOT_ENOUGH_RESOURCES: "Pattern" = re.compile(
        r"not enough (?P<resource_type>([\w\s\-\._]+)) available"
    )


__all__ = [
    "add_node_reservation",
    "add_network_reservation",
    "add_fip_reservation",
    "add_device_reservation",
    "get_node_reservation",
    "get_device_reservation",
    "get_reserved_floating_ips",
    "lease_duration",
    "get_lease",
    "get_lease_id",
    "create_lease",
    "delete_lease",
    "wait_for_active",
]

BLAZAR_TIME_FORMAT = "%Y-%m-%d %H:%M"
NODE_TYPES = {
    "compute_skylake",
    "compute_haswell_ib",
    "compute_cascadelake",
    "compute_cascadelake_r",
    "storage",
    "storage_hierarchy",
    "gpu_p100",
    "gpu_p100_nvlink",
    "gpu_k80",
    "gpu_m40",
    "fpga",
    "lowpower_xeon",
    "atom",
    "arm64",
}
DEFAULT_NODE_TYPE = "compute_skylake"
DEFAULT_LEASE_LENGTH = timedelta(days=1)
DEFAULT_NETWORK_RESOURCE_PROPERTIES = ["==", "$physical_network", "physnet1"]


def lease_create_args(
    neutronclient,
    name=None,
    start="now",
    end=None,
    length=None,
    nodes=1,
    node_resource_properties=None,
    fips=0,
    networks=0,
    network_resource_properties=DEFAULT_NETWORK_RESOURCE_PROPERTIES,
):
    """
    Generates the nested object that needs to be sent to the Blazar client
    to create the lease. Provides useful defaults for Chameleon.

    :param str name: name of lease. If ``None``, generates a random name.
    :param str/datetime start: when to start lease as a
        :py:class:`datetime.datetime` object, or if the string ``'now'``,
        starts in about a minute.
    :param length: length of time as a :py:class:`datetime.timedelta` object or
        number of seconds as a number. Defaults to 1 day.
    :param datetime.datetime end: when to end the lease. Provide only this or
        `length`, not both.
    :param int nodes: number of nodes to reserve.
    :param resource_properties: object that is JSON-encoded and sent as the
        ``resource_properties`` value to Blazar. Commonly used to specify
        node types.
    """
    if start == "now":
        start = utcnow() + timedelta(seconds=70)

    if length is None and end is None:
        length = DEFAULT_LEASE_LENGTH
    elif length is not None and end is not None:
        raise ValueError("provide either 'length' or 'end', not both")

    if end is None:
        if isinstance(length, numbers.Number):
            length = timedelta(seconds=length)
        end = start + length

    reservations = []

    if nodes > 0:
        if node_resource_properties:
            node_resource_properties = json.dumps(node_resource_properties)

        reservations += [
            {
                "resource_type": "physical:host",
                "resource_properties": node_resource_properties or "",
                "hypervisor_properties": "",
                "min": str(nodes),
                "max": str(nodes),
            }
        ]

    if fips > 0:
        reservations += [
            {
                "resource_type": "virtual:floatingip",
                "network_id": get_network_id(PUBLIC_NETWORK),
                "amount": fips,
            }
        ]

    if networks > 0:
        if network_resource_properties:
            network_resource_properties = json.dumps(network_resource_properties)

        reservations += [
            {
                "resource_type": "network",
                "resource_properties": network_resource_properties or "",
                "network_name": f"{name}-net{idx}",
            }
            for idx in range(networks)
        ]

    return {
        "name": name,
        "start": start.strftime(BLAZAR_TIME_FORMAT),
        "end": end.strftime(BLAZAR_TIME_FORMAT),
        "reservations": reservations,
        "events": [],
    }


def lease_create_nodetype(*args, **kwargs):
    """
    Wrapper for :py:func:`lease_create_args` that adds the
    ``resource_properties`` payload to specify node type.

    :param str node_type: Node type to filter by, ``compute_skylake``, et al.
    :raises ValueError: if there is no `node_type` named argument.
    """
    try:
        node_type = kwargs.pop("node_type")
    except KeyError:
        raise ValueError("no node_type specified")
    if node_type not in NODE_TYPES:
        print('warning: unknown node_type ("{}")'.format(node_type), file=sys.stderr)
        # raise ValueError('unknown node_type ("{}")'.format(node_type))
    kwargs["node_resource_properties"] = ["==", "$node_type", node_type]
    return lease_create_args(*args, **kwargs)


class Lease(object):
    """
    Creates and manages a lease, optionally with a context manager (``with``).

    .. code-block:: python

        with Lease(session, node_type='compute_skylake') as lease:
            instance = lease.create_server()
            ...

    When using the context manager, on entering it will wait for the lease
    to launch, then on exiting it will delete the lease, which in-turn
    also deletes the instances launched with it.

    :param keystone_session: session object
    :param bool sequester: If the context manager catches that an instance
        failed to start, it will not delete the lease, but rather extend it
        and rename it with the ID of the instance that failed.
    :param bool _no_clean: Don't delete the lease at the end of a context
        manager
    :param kwargs: Parameters passed through to
        :py:func:`lease_create_nodetype` and in turn
        :py:func:`lease_create_args`
    """

    def __init__(self, **kwargs):
        kwargs.setdefault("session", session())

        self.session = kwargs.pop("session")
        self.blazar = blazar(session=self.session)
        self.neutron = neutron(session=self.session)

        self.lease = None

        self._servers = {}

        self._sequester = kwargs.pop("sequester", False)

        kwargs.setdefault("_preexisting", False)
        self._preexisting = kwargs.pop("_preexisting")

        kwargs.setdefault("_no_clean", False)
        self._noclean = kwargs.pop("_no_clean")

        prefix = kwargs.pop("prefix", "")
        rand = random_base32(6)
        self.prefix = f"{prefix}-{rand}" if prefix else rand

        kwargs.setdefault("name", self.prefix)

        if self._preexisting:
            self.id = kwargs["_id"]
            self.refresh()
        else:
            kwargs.setdefault("node_type", DEFAULT_NODE_TYPE)
            self._lease_kwargs = lease_create_nodetype(self.neutron, **kwargs)
            self.lease = self.blazar.lease.create(**self._lease_kwargs)
            self.id = self.lease["id"]

        self.name = self.lease["name"]
        self.reservations = self.lease["reservations"]

    @classmethod
    def from_existing(cls, id):
        """
        Attach to an existing lease by ID. When using in conjunction with the
        context manager, it will *not* delete the lease at the end.
        """
        return cls(_preexisting=True, _id=id)

    def __repr__(self):
        region = self.session.region_name
        return "<{} '{}' on {} ({})>".format(
            self.__class__.__name__, self.name, region, self.id
        )

    def __enter__(self):
        if self.lease is None:
            # don't support reuse in multiple with's.
            raise RuntimeError("Lease context manager not reentrant")
        self.wait()
        return self

    def __exit__(self, exc_type, exc, exc_tb):
        if exc is not None and self._noclean:
            print("Lease existing uncleanly (noclean = True).")
            return

        if isinstance(exc, ServerError) and self._sequester:
            print("Instance failed to start, sequestering lease")
            self.blazar.lease.update(
                lease_id=self.id,
                name="sequester-error-instance-{}".format(exc.server.id),
                prolong_for="6d",
            )
            return

        # if lease exists, delete instances
        current_lease = self.blazar.lease.get(self.id)
        if current_lease:
            for server in self.servers:
                server.delete()

        if not self._preexisting:
            # don't auto-delete pre-existing leases
            self.delete()

    def refresh(self):
        """Updates the lease data"""
        self.lease = self.blazar.lease.get(self.id)

    @property
    def node_reservation(self):
        return next(
            iter(
                [
                    r["id"]
                    for r in (self.reservations or [])
                    if r["resource_type"] == "physical:host"
                ]
            ),
            None,
        )

    @property
    def status(self):
        """Refreshes and returns the status of the lease."""
        self.refresh()
        # NOTE(priteau): Temporary compatibility with old and new lease status
        if self.lease.get("action") is not None:
            return self.lease["action"], self.lease["status"]
        else:
            return self.lease["status"]

    @property
    def ready(self):
        """Returns True if the lease has started."""
        # NOTE(priteau): Temporary compatibility with old and new lease status
        if self.lease.get("action") is not None:
            return self.status == ("START", "COMPLETE")
        else:
            return self.status == "ACTIVE"

    @property
    def servers(self):
        return self._servers.values()

    @property
    def binding(self):
        return {
            key: {
                "address": value.ip,
                "auth": {
                    "user": "cc",
                    "private_key": get_from_context("keypair_private_key"),
                },
            }
            for key, value in self._servers.items()
        }

    def wait(self):
        """Blocks for up to 150 seconds, waiting for the lease to be ready.
        Raises a RuntimeError if it times out."""
        for _ in range(15):
            time.sleep(10)
            if self.ready:
                break
        else:
            raise RuntimeError("timeout, lease failed to start")

    def delete(self):
        """Deletes the lease"""
        self.blazar.lease.delete(self.id)
        self.lease = None

    def create_server(self, *server_args, **server_kwargs):
        """Generates instances using the resource of the lease. Arguments
        are passed to :py:class:`ccmanage.server.Server` and returns same
        object."""
        server_kwargs.setdefault("lease", self)
        server_name = server_kwargs.pop("name", len(self.servers))
        server_kwargs.setdefault("name", f"{self.prefix}-{server_name}")
        server = Server(*server_args, **server_kwargs)
        self._servers[server_name] = server
        return server


def add_node_reservation(
    reservation_list,
    count=1,
    resource_properties=[],
    node_type=None,
    architecture=None,
):
    """Add a node reservation to a reservation list.

    Args:
        reservation_list (list[dict]): The list of reservations to add to.
            The list will be extended in-place.
        count (int): The number of nodes of the given type to request.
            (Default 1).
        resource_properties (list): A list of resource property constraints. These take
            the form [<operation>, <search_key>, <search_value>], e.g.::

              ["==", "$node_type", "some-node-type"]: filter the reservation to only
                nodes with a `node_type` matching "some-node-type".
              [">", "$architecture.smt_size", 40]: filter to nodes having more than 40
                (hyperthread) cores.

        node_type (str): The node type to request. If None, the reservation will not
            target any particular node type. If `resource_properties` is defined, the
            node type constraint is added to the existing property constraints.
        architecture (str): The node architecture to request. If `resource_properties`
            is defined, the architecture constraint is added to the existing property
            constraints.
    """
    user_constraints = (resource_properties or []).copy()
    extra_constraints = []
    if node_type:
        extra_constraints.append(["==", "$node_type", node_type])
    if architecture:
        extra_constraints.append(["==", "$architecture.platform_type", architecture])

    if user_constraints:
        if user_constraints[0] == "and":
            # Already a compound constraint
            resource_properties = user_constraints + extra_constraints
        else:
            resource_properties = ["and", user_constraints] + extra_constraints
    else:
        if len(extra_constraints) < 2:
            # Possibly a compount constraint if multiple kwarg helpers used
            resource_properties = extra_constraints[0] if extra_constraints else []
        else:
            resource_properties = ["and"] + extra_constraints

    reservation_list.append(
        {
            "resource_type": "physical:host",
            "resource_properties": json.dumps(resource_properties),
            "hypervisor_properties": "",
            "min": count,
            "max": count,
        }
    )


def get_node_reservation(
    lease_ref, count=None, resource_properties=None, node_type=None, architecture=None
):
    """Retrieve a reservation ID for a node reservation.

    The reservation ID is useful to have when launching bare metal instances.

    Args:
        lease_ref (str): The ID or name of the lease.
        count (int): An optional count of nodes the desired reservation was
            made for. Use this if you have multiple reservations under a lease.
        resource_properties (list): An optional set of resource property constraints
            the desired reservation was made under. Use this if you have multiple
            reservations under a lease.
        node_type (str): An optional node type the desired reservation was
            made for. Use this if you have multiple reservations under a lease.
        architecture (str): An optional node architecture the desired reservation was
            made for. Use this if you have multiple reservations under a lease.

    Returns:
        The ID of the reservation, if found.

    Raises:
        ValueError: If no reservation was found, or multiple were found.
    """

    def _find_node_reservation(res):
        if res.get("resource_type") != "physical:host":
            return False
        if count is not None and not all(
            int(res.get(key)) == count for key in ["min_count", "max_count"]
        ):
            return False
        rp = res.get("resource_properties")
        if node_type is not None and node_type not in rp:
            return False
        if architecture is not None and architecture not in rp:
            return False
        if resource_properties is not None and json.dumps(rp) != resource_properties:
            return False
        return True

    res = _reservation_matching(lease_ref, _find_node_reservation)
    return res["id"]


def get_device_reservation(lease_ref, count=None, machine_name=None, device_model=None, device_name=None):
    """Retrieve a reservation ID for a device reservation.

    The reservation ID is useful to have when requesting containers.

    Args:
        lease_ref (str): The ID or name of the lease.
        count (int): An optional count of devices the desired reservation was
            made for. Use this if you have multiple reservations under a lease.
        machine_name (str): An optional device machine name the desired reservation
            was made for. Use this if you have multiple reservations under a lease.
        device_model (str): An optional device model the desired reservation was
            made for. Use this if you have multiple reservations under a lease.
        device_name (str): An optional device name the desired reservation was
            made for. Use this if you have multiple reservations under a lease.

    Returns:
        The ID of the reservation, if found.

    Raises:
        ValueError: If no reservation was found, or multiple were found.
    """

    def _find_device_reservation(res):
        if res.get("resource_type") != "device":
            return False
        # FIXME(jason): Blazar's device plugin uses "min" and "max", but the
        # standard seems to be "min_count" and "max_count"; this should be fixed in
        # Blazar's device plugin.
        if count is not None and not all(
            (key not in res) or int(res.get(key)) == count
            for key in ["min_count", "max_count", "min", "max"]
        ):
            return False
        resource_properties = res.get("resource_properties")
        if machine_name is not None and machine_name not in resource_properties:
            return False
        if device_model is not None and device_model not in resource_properties:
            return False
        if device_name is not None and device_name not in resource_properties:
            return False
        return True

    res = _reservation_matching(lease_ref, _find_device_reservation)
    return res["id"]


def get_reserved_floating_ips(lease_ref) -> "list[str]":
    """Get a list of Floating IP addresses reserved in a lease.

    Args:
        lease_ref (str): The ID or name of the lease.

    Returns:
        A list of all reserved Floating IP addresses, if any were reserved.
    """

    def _find_fip_reservation(res):
        return res.get("resource_type") == "virtual:floatingip"

    res = _reservation_matching(lease_ref, _find_fip_reservation, multiple=True)
    fips = list_floating_ips()
    return [
        fip["floating_ip_address"]
        for fip in fips
        if any(f"reservation:{r['id']}" in fip["tags"] for r in res)
    ]


def _reservation_matching(lease_ref, match_fn, multiple=False):
    lease = get_lease(lease_ref)
    reservations = lease.get("reservations", [])
    if isinstance(reservations, str):
        LOG.info("Blazar returned nested JSON structure, unpacking.")
        try:
            reservations = json.loads(reservations)
        except Exception:
            pass

    matches = [r for r in reservations if match_fn(r)]

    if not matches:
        raise ValueError("No matching reservation found")

    if multiple:
        return matches
    else:
        if len(matches) > 1:
            raise ValueError("Multiple matching reservations found")
        return matches[0]


def add_network_reservation(
    reservation_list,
    network_name,
    of_controller_ip=None,
    of_controller_port=None,
    vswitch_name=None,
    physical_network="physnet1",
):
    """Add a network reservation to a reservation list.

    Args:
        reservation_list (list[dict]): The list of reservations to add to.
            The list will be extended in-place.
        network_name (str): The name of the network to create when the
            reservation starts.
        of_controller_ip (str): The OpenFlow controller IP, if the network
            should be controlled by an external controller.
        of_controller_port (int): The OpenFlow controller port.
        vswitch_name (str): The name of the virtual switch associated with
            this network. See `the virtual forwarding context documentation
            <https://chameleoncloud.readthedocs.io/en/latest/technical/networks/networks_sdn.html#corsa-dp2000-virtual-forwarding-contexts-network-layout-and-advanced-features>`_
            for more details.
        physical_network (str): The physical provider network to reserve from.
            This only needs to be changed if you are reserving a `stitchable
            network <https://chameleoncloud.readthedocs.io/en/latest/technical/networks/networks_stitching.html>`_.
            (Default "physnet1").
    """
    desc_parts = []
    if of_controller_ip and of_controller_port:
        desc_parts.append(f"OFController={of_controller_ip}:{of_controller_port}")
    if vswitch_name:
        desc_parts.append(f"VSwitchName={vswitch_name}")

    reservation_list.append(
        {
            "resource_type": "network",
            "network_name": network_name,
            "network_description": ",".join(desc_parts),
            "resource_properties": json.dumps(
                ["==", "$physical_network", physical_network]
            ),
            "network_properties": "",
        }
    )


def add_fip_reservation(reservation_list, count=1):
    """Add a floating IP reservation to a reservation list.

    Args:
        reservation_list (list[dict]): The list of reservations to add to.
            The list will be extended in-place.
        count (int): The number of floating IPs to reserve.
    """
    reservation_list.append(
        {
            "resource_type": "virtual:floatingip",
            "network_id": get_network_id(PUBLIC_NETWORK),
            "amount": count,
        }
    )


def add_device_reservation(
    reservation_list, count=1, machine_name=None, device_model=None, device_name=None 
):
    """Add an IoT/edge device reservation to a reservation list.

    Args:
        reservation_list (list[dict]): The list of reservations to add to.
        count (int): The number of devices to request.
        machine_name (str): The device machine name to reserve. This should match
            a "machine_name" property of the devices registered in Blazar. This
            is the easiest way to reserve a particular device type, e.g.
            "raspberrypi4-64".
        device_model (str): The model of device to reserve. This should match
            a "model" property of the devices registered in Blazar.
        device_name (str): The name of a specific device to reserve. If this
            is provided in conjunction with ``count`` or other constraints,
            an error will be raised, as there is only 1 possible device that
            can match this criteria, because devices have unique names.

    Raises:
        ValueError: If ``device_name`` is provided, but ``count`` is greater
            than 1, or some other constraint is present.
    """
    reservation = {
        "resource_type": "device",
        "min": count,
        "max": count,
    }
    resource_properties = []
    if device_name:
        if count > 1:
            raise ValueError(
                "Cannot reserve multiple devices if device_name is a constraint."
            )
        resource_properties.append(["==", "$name", device_name])
    if machine_name:
        resource_properties.append(["==", "$machine_name", machine_name]) 
    if device_model:
        resource_properties.append(["==", "$model", device_model])

    if len(resource_properties) == 1:
        resource_properties = resource_properties[0]
    elif resource_properties:
        resource_properties.insert(0, "and")

    reservation["resource_properties"] = json.dumps(resource_properties)
    reservation_list.append(reservation)


def lease_duration(days=1, hours=0):
    """Compute the start and end dates for a lease given its desired duration.

    When providing both ``days`` and ``hours``, the duration is summed. So,
    the following would be a lease for one and a half days:

    .. code-block:: python

       start_date, end_date = lease_duration(days=1, hours=12)

    Args:
        days (int): The number of days the lease should be for.
        hours (int): The number of hours the lease should be for.
    """
    now = utcnow()
    # Start one minute into future to avoid Blazar thinking lease is in past
    # due to rounding to closest minute.
    start_date = (now + timedelta(minutes=1)).strftime(BLAZAR_TIME_FORMAT)
    end_date = (now + timedelta(days=days, hours=hours)).strftime(BLAZAR_TIME_FORMAT)
    return start_date, end_date


#########
# Leases
#########


def get_lease(ref) -> dict:
    """Get a lease by its ID or name.

    Args:
        ref (str): The ID or name of the lease.

    Returns:
        The lease matching the ID or name.
    """
    try:
        return blazar().lease.get(ref)
    except BlazarClientException as err:
        # Blazar's exception class is a bit odd and stores the actual code
        # in 'kwargs'. The 'code' attribute on the exception is just the default
        # code. Prefer to use .kwargs['code'] if present, fall back to .code
        code = getattr(err, "kwargs", {}).get("code", getattr(err, "code", None))
        if code == 404:
            return blazar().lease.get(get_lease_id(ref))
        else:
            raise


def get_lease_id(lease_name) -> str:
    """Look up a lease's ID from its name.

    Args:
        name (str): The name of the lease.

    Returns:
        The ID of the found lease.

    Raises:
        ValueError: If the lease could not be found, or if multiple leases were
            found with the same name.
    """
    matching = [l for l in blazar().lease.list() if l["name"] == lease_name]
    if not matching:
        raise ValueError(f"No leases found for name {lease_name}")
    elif len(matching) > 1:
        raise ValueError(f"Multiple leases found for name {lease_name}")
    return matching[0]["id"]


def create_lease(lease_name, reservations=[], start_date=None, end_date=None):
    """Create a new lease with some requested reservations.

    Args:
        lease_name (str): The name to give the new lease.
        reservations (list[dict]): The reservations to request with the lease.
        start_date (datetime): The start date of the lease. (Defaults to now.)
        end_date (datetime): The end date of the lease. (Defaults to 1 day from
            the lease start date.)

    Returns:
        The created lease representation.
    """
    if not (start_date or end_date):
        start_date, end_date = lease_duration(days=1)
    elif not end_date:
        end_date = start_date + timedelta(days=1)
    elif not start_date:
        start_date = utcnow()

    if not reservations:
        raise ValueError("No reservations provided.")

    try:
        return blazar().lease.create(
            name=lease_name,
            start=start_date,
            end=end_date,
            reservations=reservations,
            events=[],
        )
    except BlazarClientException as ex:
        msg: "str" = ex.args[0]
        msg = msg.lower()

        match = ErrorParsers.NOT_ENOUGH_RESOURCES.match(msg)
        if match:
            LOG.error(
                f"There were not enough unreserved {match.group('resource_type')} "
                "to satisfy your request."
            )
        else:
            LOG.error(msg)


def delete_lease(ref):
    """Delete the lease.

    Args:
        ref (str): The name or ID of the lease.
    """
    lease = get_lease(ref)
    lease_id = lease["id"]
    blazar().lease.delete(lease_id)
    print(f"Deleted lease with id {lease_id}")


def wait_for_active(ref):
    """Wait for the lease to become active.

    This function will wait for 2.5 minutes, which is a somewhat arbitrary
    amount of time.

    Args:
        ref (str): The name or ID of the lease.

    Returns:
        The lease in ACTIVE state.

    Raises:
        TimeoutError: If the lease fails to become active within the timeout.
    """
    for _ in range(15):
        lease = get_lease(ref)
        status = lease["status"]
        if status == "ACTIVE":
            return lease
        elif status == "ERROR":
            raise RuntimeError("Lease went into ERROR state")
        time.sleep(10)
    raise TimeoutError("Lease failed to start")



from .clients import glance

from glanceclient.exc import NotFound

__all__ = [
    'get_image',
    'get_image_id',
    'list_images',
]


def get_image(ref):
    """Get an image by its ID or name.

    Args:
        ref (str): The ID or name of the image.

    Returns:
        The image matching the ID or name.

    Raises:
        NotFound: If the image could not be found.
    """
    try:
        return glance().images.get(ref)
    except NotFound:
        return glance().images.get(get_image_id(ref))


def get_image_id(name):
    """Look up an image's ID from its name.

    Args:
        name (str): The name of the image.

    Returns:
        The ID of the found image.

    Raises:
        ValueError: If the image could not be found, or if multiple images
            matched the name.
    """
    images = list(glance().images.list(filters={'name': name}))
    if not images:
        raise ValueError(f'No images found matching name "{name}"')
    elif len(images) > 1:
        raise ValueError(f'Multiple images found matching name "{name}"')
    return images[0].id


def list_images():
    """List all images under the current project.

    Returns:
        All images associated with the current project.
    """
    return list(glance().images.list())



from datetime import datetime

import pytest

@pytest.fixture()
def now():
    return datetime(2021, 1, 1, 0, 0, 0, 0)


def example_reserve_node():
    """Reserve a bare metal node.

    Multiple nodes can be reserved at once by changing the `count` variable.
    This example makes a reservation for the "compute_skylake" node type.
    [See here](https://chameleoncloud.readthedocs.io/en/latest/technical/reservations.html#chameleon-node-types)
    for a complete list of node types available currently.

    <div class="alert alert-info">

    **Functions used in this example:**

    * [add_node_reservation](../modules/lease.html#chi.lease.add_node_reservation)
    * [lease_duration](../modules/lease.html#chi.lease.lease_duration)
    * [create_lease](../modules/lease.html#chi.lease.create_lease)

    </div>

    """
    from chi.lease import lease_duration, add_node_reservation, create_lease

    lease_name = "myLease"
    node_type = "compute_skylake"
    start_date, end_date = lease_duration(days=1)

    # Build list of reservations (in this case there is only one reservation)
    reservations = []
    add_node_reservation(reservations, count=1, node_type=node_type)
    # Create the lease
    lease = create_lease(lease_name, reservations, start_date=start_date,
                         end_date=end_date)


def test_example_reserve_node(mocker, now):
    mocker.patch('chi.lease.utcnow', return_value=now)
    blazar = mocker.patch('chi.lease.blazar')()

    example_reserve_node()

    blazar.lease.create.assert_called_once_with(
        name='myLease',
        start='2021-01-01 00:01',
        end='2021-01-02 00:00',
        events=[],
        reservations=[{
            'resource_type': 'physical:host',
            'hypervisor_properties': '', 'max': 1, 'min': 1,
            'resource_properties': '["==", "$node_type", "compute_skylake"]',
        }]
    )


def example_reserve_network():
    """Reserve a VLAN segment.

    This example illustrates how to reserve an isolated VLAN in order to
    ensure your network experiment is not subject to cross-traffic from other
    experimenters.

    This is also how you reserve
    [stitchable VLANs](https://chameleoncloud.readthedocs.io/en/latest/technical/networks/networks_stitching.html)
    provided through ExoGENI. For these VLANs, you must set `physical_network`
    to "exogeni".

    <div class="alert alert-info">

    **Functions used in this example:**

    * [add_network_reservation](../modules/lease.html#chi.lease.add_network_reservation)
    * [lease_duration](../modules/lease.html#chi.lease.lease_duration)
    * [create_lease](../modules/lease.html#chi.lease.create_lease)

    </div>

    """
    from chi.lease import lease_duration, add_network_reservation, create_lease

    lease_name = "myLease"
    network_name = f"{lease_name}Network"
    of_controller_ip = None
    of_controller_port = None
    vswitch_name = None
    physical_network = "physnet1"
    start_date, end_date = lease_duration(days=1)

    # Build list of reservations (in this case there is only one reservation)
    reservations = []
    add_network_reservation(reservations,
                            network_name=network_name,
                            of_controller_ip=of_controller_ip,
                            of_controller_port=of_controller_port,
                            vswitch_name=vswitch_name,
                            physical_network=physical_network)

    # Create the lease
    lease = create_lease(lease_name, reservations, start_date=start_date,
                         end_date=end_date)


def test_example_reserve_network(mocker, now):
    mocker.patch('chi.lease.utcnow', return_value=now)
    blazar = mocker.patch('chi.lease.blazar')()

    example_reserve_network()

    blazar.lease.create.assert_called_once_with(
        name='myLease',
        start='2021-01-01 00:01',
        end='2021-01-02 00:00',
        events=[],
        reservations=[{
            'resource_type': 'network',
            'network_name': 'myLeaseNetwork',
            'network_description': '',
            'network_properties': '',
            'resource_properties': '["==", "$physical_network", "physnet1"]',
        }]
    )


def example_reserve_floating_ip():
    """Reserve a floating IP.

    While it's possible to allocate a floating IP ad hoc from Chameleon most
    of the time, there are a limited amount of IPs and they are sometimes
    exhausted. You can reserve a floating IP to ensure you have access to one
    to attach to your experimental nodes to allow, e.g., external SSH
    connectivity.

    [See here](https://www.chameleoncloud.org/blog/2019/02/27/save-planet-use-fewer-ips/)
    for some tips on how to make the most out of a single floating IP, which
    can help you avoid excessive charges.

    <div class="alert alert-info">

    **Functions used in this example:**

    * [add_fip_reservation](../modules/lease.html#chi.lease.add_fip_reservation)
    * [lease_duration](../modules/lease.html#chi.lease.lease_duration)
    * [create_lease](../modules/lease.html#chi.lease.create_lease)

    </div>

    """
    from chi.lease import lease_duration, add_fip_reservation, create_lease

    lease_name = "myLease"
    start_date, end_date = lease_duration(days=1)

    # Build list of reservations (in this case there is only one reservation)
    reservation_list = []
    add_fip_reservation(reservation_list, count=1)

    # Create the lease
    lease = create_lease(lease_name, reservation_list, start_date=start_date,
                         end_date=end_date)


def test_example_reserve_floating_ip(mocker, now):
    mocker.patch('chi.lease.utcnow', return_value=now)
    blazar = mocker.patch('chi.lease.blazar')()
    mocker.patch('chi.lease.get_network_id', return_value='public-net-id')

    example_reserve_floating_ip()

    blazar.lease.create.assert_called_once_with(
        name='myLease',
        start='2021-01-01 00:01',
        end='2021-01-02 00:00',
        events=[],
        reservations=[{
            'resource_type': 'virtual:floatingip',
            'amount': 1,
            'network_id': 'public-net-id',
        }]
    )


def example_reserve_multiple_resources():
    """Reserve multiple types of resources in a single lease.

    <div class="alert alert-info">

    **Functions used in this example:**

    * [add_node_reservation](../modules/lease.html#chi.lease.add_node_reservation)
    * [add_network_reservation](../modules/lease.html#chi.lease.add_network_reservation)
    * [add_fip_reservation](../modules/lease.html#chi.lease.add_fip_reservation)
    * [lease_duration](../modules/lease.html#chi.lease.lease_duration)
    * [create_lease](../modules/lease.html#chi.lease.create_lease)

    </div>

    """
    from chi.lease import (
        lease_duration, add_node_reservation, add_network_reservation,
        add_fip_reservation, create_lease)

    lease_name = "myLease"
    start_date, end_date = lease_duration(days=1)

    # Build list of reservations
    reservations = []
    add_node_reservation(reservations, count=1, node_type="compute_skylake")
    add_network_reservation(reservations, network_name=f"{lease_name}Network")
    add_fip_reservation(reservations, count=1)

    # Create the lease
    lease = create_lease(lease_name, reservations, start_date=start_date,
                         end_date=end_date)


def test_example_reserve_multiple_resources(mocker, now):
    mocker.patch('chi.lease.utcnow', return_value=now)
    blazar = mocker.patch('chi.lease.blazar')()
    mocker.patch('chi.lease.get_network_id', return_value='public-net-id')

    example_reserve_multiple_resources()

    blazar.lease.create.assert_called_once_with(
        name='myLease',
        start='2021-01-01 00:01',
        end='2021-01-02 00:00',
        events=[],
        reservations=[{
            'resource_type': 'physical:host',
            'hypervisor_properties': '', 'max': 1, 'min': 1,
            'resource_properties': '["==", "$node_type", "compute_skylake"]',
        }, {
            'resource_type': 'network',
            'network_name': 'myLeaseNetwork',
            'network_description': '',
            'network_properties': '',
            'resource_properties': '["==", "$physical_network", "physnet1"]',
        }, {
            'resource_type': 'virtual:floatingip',
            'amount': 1,
            'network_id': 'public-net-id',
        }]
    )



from neutronclient.common.exceptions import NotFound
import pytest

from chi.network import get_network, get_subnet, get_router, get_floating_ip


def fake_network():
    return {
        "id": "network-id",
        "name": "network-name",
        "project_id": "network-project-id",
    }


def fake_subnet():
    return {
        "id": "subnet-id",
        "name": "subnet-name",
        "network_id": "network-id",
    }


def fake_router():
    return {
        "id": "router-id",
        "name": "router-name",
        "project_id": "router-project-id",
    }


def fake_floating_ip():
    return {
        "id": "floatingip-id",
        "name": "floatingip-name",
        "project_id": "floatingip-project-id",
    }


@pytest.mark.parametrize("test_fn,neutron_resource,expected", [
    pytest.param(get_network, "network", fake_network(), id="network"),
    pytest.param(get_subnet, "subnet", fake_subnet(), id="subnet"),
    pytest.param(get_router, "router", fake_router(), id="router"),
])
def test_get_resource_by_id(mocker, test_fn, neutron_resource, expected):
    neutron = mocker.patch("chi.network.neutron")()

    neutron_show = getattr(neutron, f"show_{neutron_resource}")
    neutron_show.return_value = {neutron_resource: expected}

    search_id = f"{neutron_resource}-id"
    assert test_fn(search_id) == expected
    neutron_show.assert_called_once_with(search_id)


@pytest.mark.parametrize("test_fn,neutron_resource,expected", [
    pytest.param(get_network, "network", fake_network(), id="network"),
    pytest.param(get_subnet, "subnet", fake_subnet(), id="subnet"),
    pytest.param(get_router, "router", fake_router(), id="router"),
])
def test_get_resource_by_name(mocker, test_fn, neutron_resource, expected):
    neutron = mocker.patch("chi.network.neutron")()

    search_id = f"{neutron_resource}-id"
    search_name = f"{neutron_resource}-name"

    def _show(ref):
        if ref == search_name:
            raise NotFound
        elif ref == search_id:
            return {neutron_resource: expected}
        else:
            raise ValueError(f"Unexpected ref {ref}")

    neutron_show = getattr(neutron, f"show_{neutron_resource}")
    neutron_list = getattr(neutron, f"list_{neutron_resource}s")
    neutron_show.side_effect = _show
    neutron_list.return_value = {f"{neutron_resource}s": [expected]}

    assert test_fn(search_name) == expected
    # In this case we expect 1 call for the name, which will fail, then a call
    # to list resources to find the ID, then a call to look up by ID.
    neutron_show.assert_any_call(search_name)
    neutron_list.assert_called_once()
    neutron_show.assert_any_call(search_id)
    assert neutron_show.call_count == 2
